@@ -38,11 +38,13 @@
 #include <linux/fs.h>
 
 #include "nls.h"
+#include "xalloc.h"
 #include "strutils.h"
 #include "c.h"
 #include "closestream.h"
 #include "pathnames.h"
 #include "sysfs.h"
+#include "optutils.h"
 
 #include <libmount.h>
 
@@ -60,36 +62,48 @@ struct fstrim_control {
 	struct fstrim_range range;
 
 	unsigned int verbose : 1,
-		     fstab   : 1,
+		     quiet_unsupp : 1,
 		     dryrun : 1;
 };
 
+static int is_directory(const char *path, int silent)
+{
+	struct stat sb;
+
+	if (stat(path, &sb) == -1) {
+		if (!silent)
+			warn(_("stat of %s failed"), path);
+		return 0;
+	}
+	if (!S_ISDIR(sb.st_mode)) {
+		if (!silent)
+			warnx(_("%s: not a directory"), path);
+		return 0;
+	}
+	return 1;
+}
+
 /* returns: 0 = success, 1 = unsupported, < 0 = error */
 static int fstrim_filesystem(struct fstrim_control *ctl, const char *path, const char *devname)
 {
-	int fd, rc;
-	struct stat sb;
+	int fd = -1, rc;
 	struct fstrim_range range;
+	char *rpath = realpath(path, NULL);
 
+	if (!rpath) {
+		warn(_("cannot get realpath: %s"), path);
+		rc = -errno;
+		goto done;
+	}
 	/* kernel modifies the range */
 	memcpy(&range, &ctl->range, sizeof(range));
 
-	fd = open(path, O_RDONLY);
+	fd = open(rpath, O_RDONLY);
 	if (fd < 0) {
 		warn(_("cannot open %s"), path);
 		rc = -errno;
 		goto done;
 	}
-	if (fstat(fd, &sb) == -1) {
-		warn(_("stat of %s failed"), path);
-		rc = -errno;
-		goto done;
-	}
-	if (!S_ISDIR(sb.st_mode)) {
-		warnx(_("%s: not a directory"), path);
-		rc = -EINVAL;
-		goto done;
-	}
 
 	if (ctl->dryrun) {
 		if (devname)
@@ -102,9 +116,16 @@ static int fstrim_filesystem(struct fstrim_control *ctl, const char *path, const
 
 	errno = 0;
 	if (ioctl(fd, FITRIM, &range)) {
-		rc = errno == EOPNOTSUPP || errno == ENOTTY ? 1 : -errno;
-
-		if (rc != 1)
+		switch (errno) {
+		case EBADF:
+		case ENOTTY:
+		case EOPNOTSUPP:
+			rc = 1;
+			break;
+		default:
+			rc = -errno;
+		}
+		if (rc < 0)
 			warn(_("%s: FITRIM ioctl failed"), path);
 		goto done;
 	}
@@ -129,6 +150,7 @@ static int fstrim_filesystem(struct fstrim_control *ctl, const char *path, const
 done:
 	if (fd >= 0)
 		close(fd);
+	free(rpath);
 	return rc;
 }
 
@@ -137,7 +159,7 @@ static int has_discard(const char *devname, struct path_cxt **wholedisk)
 	struct path_cxt *pc = NULL;
 	uint64_t dg = 0;
 	dev_t disk = 0, dev;
-	int rc = -1;
+	int rc = -1, rdonly = 0;
 
 	dev = sysfs_devname_to_devno(devname);
 	if (!dev)
@@ -175,9 +197,11 @@ static int has_discard(const char *devname, struct path_cxt **wholedisk)
 	}
 
 	rc = ul_path_read_u64(pc, &dg, "queue/discard_granularity");
+	if (!rc)
+		ul_path_scanf(pc, "ro", "%d", &rdonly);
 
 	ul_unref_path(pc);
-	return rc == 0 && dg > 0;
+	return rc == 0 && dg > 0 && rdonly == 0;
 fail:
 	ul_unref_path(pc);
 	return 1;
@@ -205,13 +229,12 @@ static int uniq_fs_source_cmp(
 }
 
 /*
- * fstrim --all follows "mount -a" return codes:
- *
- * 0  = all success
+ * -1 = tab empty
+ *  0 = all success
  * 32 = all failed
  * 64 = some failed, some success
  */
-static int fstrim_all(struct fstrim_control *ctl)
+static int fstrim_all_from_file(struct fstrim_control *ctl, const char *filename)
 {
 	struct libmnt_fs *fs;
 	struct libmnt_iter *itr;
@@ -219,65 +242,110 @@ static int fstrim_all(struct fstrim_control *ctl)
 	struct libmnt_cache *cache = NULL;
 	struct path_cxt *wholedisk = NULL;
 	int cnt = 0, cnt_err = 0;
-	const char *filename = _PATH_PROC_MOUNTINFO;
-
-	mnt_init_debug(0);
-	ul_path_init_debug();
-
-	itr = mnt_new_iter(MNT_ITER_BACKWARD);
-	if (!itr)
-		err(MNT_EX_FAIL, _("failed to initialize libmount iterator"));
-
-	if (ctl->fstab)
-		filename = mnt_get_fstab_path();
+	int fstab = 0;
 
 	tab = mnt_new_table_from_file(filename);
 	if (!tab)
 		err(MNT_EX_FAIL, _("failed to parse %s"), filename);
 
+	if (mnt_table_is_empty(tab)) {
+		mnt_unref_table(tab);
+		return -1;
+	}
+
+	if (streq_paths(filename, "/etc/fstab"))
+		fstab = 1;
+
 	/* de-duplicate by mountpoints */
 	mnt_table_uniq_fs(tab, 0, uniq_fs_target_cmp);
 
+	if (fstab) {
+		char *rootdev = NULL;
+
+		cache = mnt_new_cache();
+		if (!cache)
+			err(MNT_EX_FAIL, _("failed to initialize libmount cache"));
+
+		/* Make sure we trim also root FS on fstab */
+		if (mnt_table_find_target(tab, "/", MNT_ITER_FORWARD) == NULL &&
+		    mnt_guess_system_root(0, cache, &rootdev) == 0) {
+
+			fs = mnt_new_fs();
+			if (!fs)
+				err(MNT_EX_FAIL, _("failed to allocate FS handler"));
+			mnt_fs_set_target(fs, "/");
+			mnt_fs_set_source(fs, rootdev);
+			mnt_fs_set_fstype(fs, "auto");
+			mnt_table_add_fs(tab, fs);
+			mnt_unref_fs(fs);
+			fs = NULL;
+		}
+		free(rootdev);
+	}
+
+	itr = mnt_new_iter(MNT_ITER_BACKWARD);
+	if (!itr)
+		err(MNT_EX_FAIL, _("failed to initialize libmount iterator"));
+
+	/* Remove useless entries and canonicalize the table */
+	while (mnt_table_next_fs(tab, itr, &fs) == 0) {
+		const char *src = mnt_fs_get_srcpath(fs),
+			   *tgt = mnt_fs_get_target(fs);
+
+		if (!tgt || mnt_fs_is_pseudofs(fs) || mnt_fs_is_netfs(fs)) {
+			mnt_table_remove_fs(tab, fs);
+			continue;
+		}
+
+		/* convert LABEL= (etc.) from fstab to paths */
+		if (!src && cache) {
+			const char *spec = mnt_fs_get_source(fs);
+
+			if (!spec) {
+				mnt_table_remove_fs(tab, fs);
+				continue;
+			}
+			src = mnt_resolve_spec(spec, cache);
+			mnt_fs_set_source(fs, src);
+		}
+
+		if (!src || *src != '/') {
+			mnt_table_remove_fs(tab, fs);
+			continue;
+		}
+	}
+
 	/* de-duplicate by source */
 	mnt_table_uniq_fs(tab, MNT_UNIQ_FORWARD, uniq_fs_source_cmp);
 
-	if (ctl->fstab) {
-		cache = mnt_new_cache();
-		if (!cache)
-			err(MNT_EX_FAIL, _("failed to initialize libmount cache"));
-	}
+	mnt_reset_iter(itr, MNT_ITER_BACKWARD);
 
+	/* Do FITRIM */
 	while (mnt_table_next_fs(tab, itr, &fs) == 0) {
 		const char *src = mnt_fs_get_srcpath(fs),
 			   *tgt = mnt_fs_get_target(fs);
 		char *path;
 		int rc = 1;
 
-		if (!tgt || mnt_fs_is_pseudofs(fs) || mnt_fs_is_netfs(fs))
-			continue;
-
-		if (!src && cache) {
-			/* convert LABEL= (etc.) from fstab to paths */
-			const char *spec = mnt_fs_get_source(fs);
-
-			if (!spec)
-				continue;
-			src = mnt_resolve_spec(spec, cache);
-		}
-
-		if (!src || *src != '/')
-			continue;
-
 		/* Is it really accessible mountpoint? Not all mountpoints are
 		 * accessible (maybe over mounted by another filesystem) */
 		path = mnt_get_mountpoint(tgt);
-		if (path && strcmp(path, tgt) == 0)
+		if (path && streq_paths(path, tgt))
 			rc = 0;
 		free(path);
 		if (rc)
 			continue;	/* overlaying mount */
 
-		if (!has_discard(src, &wholedisk))
+		/* FITRIM on read-only filesystem can fail, and it can fail */
+		if (access(tgt, W_OK) != 0) {
+			if (errno == EROFS)
+				continue;
+			if (errno == EACCES)
+				continue;
+		}
+
+		if (!is_directory(tgt, 1) ||
+		    !has_discard(src, &wholedisk))
 			continue;
 		cnt++;
 
@@ -289,13 +357,16 @@ static int fstrim_all(struct fstrim_control *ctl)
 		 * This is reason why we ignore EOPNOTSUPP and ENOTTY errors
 		 * from discard ioctl.
 		 */
-		if (fstrim_filesystem(ctl, tgt, src) < 0)
+		rc = fstrim_filesystem(ctl, tgt, src);
+		if (rc < 0)
 		       cnt_err++;
+		else if (rc == 1 && !ctl->quiet_unsupp)
+			warnx(_("%s: the discard operation is not supported"), tgt);
 	}
+	mnt_free_iter(itr);
 
 	ul_unref_path(wholedisk);
 	mnt_unref_table(tab);
-	mnt_free_iter(itr);
 	mnt_unref_cache(cache);
 
 	if (cnt && cnt == cnt_err)
@@ -306,6 +377,36 @@ static int fstrim_all(struct fstrim_control *ctl)
 	return MNT_EX_SUCCESS;
 }
 
+/*
+ * fstrim --all follows "mount -a" return codes:
+ *
+ * 0  = all success
+ * 32 = all failed
+ * 64 = some failed, some success
+ */
+static int fstrim_all(struct fstrim_control *ctl, const char *tabs)
+{
+	char *list = xstrdup(tabs);
+	char *file;
+	int rc = MNT_EX_FAIL;
+
+	mnt_init_debug(0);
+	ul_path_init_debug();
+
+	for (file = strtok(list, ":"); file; file = strtok(NULL, ":")) {
+		struct stat st;
+
+		if (stat(file, &st) < 0 || !S_ISREG(st.st_mode))
+			continue;
+
+		rc = fstrim_all_from_file(ctl, file);
+		if (rc >= 0)
+			break;	/* stop after first non-empty file */
+	}
+	free(list);
+	return rc;
+}
+
 static void __attribute__((__noreturn__)) usage(void)
 {
 	FILE *out = stdout;
@@ -317,16 +418,22 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_("Discard unused blocks on a mounted filesystem.\n"), out);
 
 	fputs(USAGE_OPTIONS, out);
-	fputs(_(" -a, --all           trim all supported mounted filesystems\n"), out);
-	fputs(_(" -A, --fstab         trim all supported mounted filesystems from /etc/fstab\n"), out);
-	fputs(_(" -o, --offset <num>  the offset in bytes to start discarding from\n"), out);
-	fputs(_(" -l, --length <num>  the number of bytes to discard\n"), out);
-	fputs(_(" -m, --minimum <num> the minimum extent length to discard\n"), out);
-	fputs(_(" -v, --verbose       print number of discarded bytes\n"), out);
-	fputs(_(" -n, --dry-run       does everything, but trim\n"), out);
+	fputs(_(" -a, --all                trim mounted filesystems\n"), out);
+	fputs(_(" -A, --fstab              trim filesystems from /etc/fstab\n"), out);
+	fputs(_(" -I, --listed-in <list>   trim filesystems listed in specified files\n"), out);
+	fputs(_(" -o, --offset <num>       the offset in bytes to start discarding from\n"), out);
+	fputs(_(" -l, --length <num>       the number of bytes to discard\n"), out);
+	fputs(_(" -m, --minimum <num>      the minimum extent length to discard\n"), out);
+	fputs(_(" -v, --verbose            print number of discarded bytes\n"), out);
+	fputs(_("     --quiet-unsupported  suppress error messages if trim unsupported\n"), out);
+	fputs(_(" -n, --dry-run            does everything, but trim\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	printf(USAGE_HELP_OPTIONS(21));
+
+	fputs(USAGE_ARGUMENTS, out);
+	printf(USAGE_ARG_SIZE(_("<num>")));
+
 	printf(USAGE_MAN_TAIL("fstrim(8)"));
 	exit(EXIT_SUCCESS);
 }
@@ -334,46 +441,61 @@ static void __attribute__((__noreturn__)) usage(void)
 int main(int argc, char **argv)
 {
 	char *path = NULL;
+	char *tabs = NULL;
 	int c, rc, all = 0;
 	struct fstrim_control ctl = {
 			.range = { .len = ULLONG_MAX }
 	};
+	enum {
+		OPT_QUIET_UNSUPP = CHAR_MAX + 1
+	};
 
 	static const struct option longopts[] = {
 	    { "all",       no_argument,       NULL, 'a' },
 	    { "fstab",     no_argument,       NULL, 'A' },
 	    { "help",      no_argument,       NULL, 'h' },
+	    { "listed-in", required_argument, NULL, 'I' },
 	    { "version",   no_argument,       NULL, 'V' },
 	    { "offset",    required_argument, NULL, 'o' },
 	    { "length",    required_argument, NULL, 'l' },
 	    { "minimum",   required_argument, NULL, 'm' },
 	    { "verbose",   no_argument,       NULL, 'v' },
+	    { "quiet-unsupported", no_argument,       NULL, OPT_QUIET_UNSUPP },
 	    { "dry-run",   no_argument,       NULL, 'n' },
 	    { NULL, 0, NULL, 0 }
 	};
 
+	static const ul_excl_t excl[] = {       /* rows and cols in ASCII order */
+		{ 'A','I','a' },
+		{ 0 }
+	};
+	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
+
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout);
+	close_stdout_atexit();
+
+	while ((c = getopt_long(argc, argv, "AahI:l:m:no:Vv", longopts, NULL)) != -1) {
+
+		err_exclusive_options(c, longopts, excl, excl_st);
 
-	while ((c = getopt_long(argc, argv, "Aahl:m:no:Vv", longopts, NULL)) != -1) {
 		switch(c) {
 		case 'A':
-			ctl.fstab = 1;
-			/* fallthrough */
+			all = 1;
+			tabs = _PATH_MNTTAB;	/* fstab */
+			break;
 		case 'a':
 			all = 1;
+			tabs = _PATH_PROC_MOUNTINFO; /* mountinfo */
+			break;
+		case 'I':
+			all = 1;
+			tabs = optarg;
 			break;
 		case 'n':
 			ctl.dryrun = 1;
 			break;
-		case 'h':
-			usage();
-			break;
-		case 'V':
-			printf(UTIL_LINUX_VERSION);
-			return EXIT_SUCCESS;
 		case 'l':
 			ctl.range.len = strtosize_or_err(optarg,
 					_("failed to parse length"));
@@ -389,9 +511,15 @@ int main(int argc, char **argv)
 		case 'v':
 			ctl.verbose = 1;
 			break;
+		case OPT_QUIET_UNSUPP:
+			ctl.quiet_unsupp = 1;
+			break;
+		case 'h':
+			usage();
+		case 'V':
+			print_version(EXIT_SUCCESS);
 		default:
 			errtryhelp(EXIT_FAILURE);
-			break;
 		}
 	}
 
@@ -407,9 +535,14 @@ int main(int argc, char **argv)
 	}
 
 	if (all)
-		return fstrim_all(&ctl);	/* MNT_EX_* codes */
+		return fstrim_all(&ctl, tabs);	/* MNT_EX_* codes */
+
+	if (!is_directory(path, 0))
+		return EXIT_FAILURE;
 
 	rc = fstrim_filesystem(&ctl, path, NULL);
+	if (rc == 1 && ctl.quiet_unsupp)
+		rc = 0;
 	if (rc == 1)
 		warnx(_("%s: the discard operation is not supported"), path);
 
